// MAKPATCH.CPP  Creates a .PAT patch file from 2 .NES ROM files

#include <stdio.h>
#include <conio.h>

long int filesize(FILE *);

int main(int argc, char *argv[])
{
	FILE *originalrom, *outputpatch, *modifiedrom; // pointers to the .NES ROMs and patch
	char *origromname, *patchname, *modfromname;
	long int origsize, modfsize, startloc, datalength, tval,
		   c, c2, sharedbytes;
	unsigned char onechar;

	// dynamically allocate 3 char arrays to store filenames
	origromname = new char [13];
	patchname = new char [13];
	modfromname = new char [13];

	if((argc<3) || (argc>4)) // if invalid number of parameters passed in
	{
		printf("\nMAKPATCH - Creates a *.PAT patch file from 2 given ROM files.\n");
		printf("Usage of the MAKPATCH command:\n\n");
		printf("      MAKPATCH <original ROM> <modified ROM> <patch>\n\n");
		printf("If no patch filename is given, the modified ROM filename will be used.\n");
		printf("Default ROM extension is *.NES, default patch extension is *.PAT.\n");
		return(0);
	}

	// First read in the original ROM filename that was passed in
	for(c=0;c<9;c++)
	{
		origromname[c]=argv[1][c];
		if((origromname[c]=='.') || (origromname[c]=='\0')) // if terminator reached
			break;
	}

	if(c==9) // if original ROM filename too long
	{
		printf("Invalid ROM #1 filename.  Aborting.\n");
		return(0);
	}

	if(origromname[c] == '.') // if there is an extension
	{
		c++;  // :)    move past '.' to get to extension

		for(c2=0;c2<4;c2++)
		{
			origromname[c+c2]=argv[1][c+c2];
			if(origromname[c+c2]=='\0') // if terminator reached
				break;
		}

		if(c2==4) // if original ROM extension too long
		{
			printf("Invalid ROM #1 extension.  Aborting.\n");
			return(0);
		}
	}
	else // else add the default .NES extension
	{
		origromname[c]   = '.';
		origromname[c+1] = 'N';
		origromname[c+2] = 'E';
		origromname[c+3] = 'S';
		origromname[c+4] = '\0';
	}

	// Now read	in the modified ROM filename that was passed in
	for(c=0;c<9;c++)
	{
		modfromname[c]=argv[2][c];
		if((modfromname[c]=='.') || (modfromname[c]=='\0')) // if terminator reached
			break;
	}

	if(c==9) // if modified ROM filename too long
	{
		printf("Invalid ROM #2 filename.  Aborting.\n");
		return(0);
	}

	if(modfromname[c] == '.') // if there is an extension
	{
		c++;  // :)    move past '.' to get to extension

		for(c2=0;c2<4;c2++)
		{
			modfromname[c+c2]=argv[2][c+c2];
			if(modfromname[c+c2]=='\0') // if terminator reached
				break;
		}

		if(c2==4) // if modified ROM extension too long
		{
			printf("Invalid ROM #2 extension.  Aborting.\n");
			return(0);
		}
	}
	else // else add the default .NES extension
	{
		modfromname[c]   = '.';
		modfromname[c+1] = 'N';
		modfromname[c+2] = 'E';
		modfromname[c+3] = 'S';
		modfromname[c+4] = '\0';
	}

	if((originalrom = fopen(origromname, "rb")) == NULL) // if source file not present
	{
		printf("Original ROM file not found.  Aborting.\n");
		return(0);
	}

	if((modifiedrom = fopen(modfromname, "rb")) == NULL) // if source file not present
	{
		printf("Modified ROM file not found.  Aborting.\n");
		return(0);
	}

	origsize = filesize(originalrom); // read size of original ROM
	modfsize = filesize(modifiedrom); // read size of modified ROM

	if(origsize != modfsize) // if they are not the same
	{
		printf("Source ROM size differs from modified ROM size.  Aborting.\n");
		return(0);
	}

	if(argc == 4) // If patch filename given
	{
		// Read in the patch filename passed in
		for(c=0;c<9;c++)
		{
			patchname[c]=argv[3][c];
			if((patchname[c]=='.') || (patchname[c]=='\0')) // if terminator reached
				break;
		}

		if(c==9) // if dest filename too long
		{
			printf("Invalid patch filename.  Aborting.\n");
			return(0);
		}

		if(patchname[c] == '.') // if there is an extension
		{
			c++;  // :)    move past '.' to get to extension

			for(c2=0;c2<4;c2++)
			{
				patchname[c+c2]=argv[3][c+c2];
				if(patchname[c+c2]=='\0') // if terminator reached
					break;
			}

			if(c2==4) // if patch extension too long
			{
				printf("Invalid patch extension.  Aborting.\n");
				return(0);
			}
		}
		else // else add the default .PAT extension
		{
			patchname[c]   = '.';
			patchname[c+1] = 'P';
			patchname[c+2] = 'A';
			patchname[c+3] = 'T';
			patchname[c+4] = '\0';
		}
	}
	else // else make a new patch with the filename of the modified ROM, and extension .PAT
	{
		for(c=0;c<8;c++)
		{
			patchname[c] = modfromname[c]; // Copy modified ROM filename to patch filename
			if(patchname[c]=='.')
				break;
		}

		// Add default .PAT extension
		patchname[c]   = '.';
		patchname[c+1] = 'P';
		patchname[c+2] = 'A';
		patchname[c+3] = 'T';
		patchname[c+4] = '\0';
	}

	outputpatch = fopen(patchname, "w+b"); // open new patch file with write enabled (if one exists, it will erase it)

	// At this point, no longer need the filenames so deallocate them
	delete[] origromname;
	delete[] modfromname;

	// NOW, we can start creating the patch data from the ROMs.
	// (not too easy) :)

	fseek(outputpatch, 0, SEEK_SET); // go to beginning of patch file

	// first write the ROM file size for later verification
	for(c=0;c<4;c++)
	{
		onechar = modfsize % 0x100; // get LSByte
		fputc(onechar, outputpatch);
		modfsize /= 0x100; // move up 1 byte
	}

	startloc=0L; // set starting ROM location at 0

	while(startloc < origsize) // while code left to examine
	{
		fseek(originalrom, startloc, SEEK_SET); // go to startloc
		fseek(modifiedrom, startloc, SEEK_SET); // in both roms

		// first look for the first byte that is different, this
		// is where the patch segment will begin
		while((startloc < origsize) && (fgetc(originalrom) == fgetc(modifiedrom)))
			startloc++;

		if(startloc >= origsize) // if EOF
			break; // finished

		fseek(originalrom, startloc, SEEK_SET); // return to startloc
		fseek(modifiedrom, startloc, SEEK_SET); // in both roms
		datalength=0; // set initial data length to 0
		sharedbytes=0; // set initial # of shared bytes to 0

		// now look to see how many bytes to include in the data segment
		while(((startloc+datalength) < origsize) && (sharedbytes < 6))
		{
			if(fgetc(originalrom) == fgetc(modifiedrom)) // if byte is same
				sharedbytes++; // tally it
			else sharedbytes=0; // else reset tally

			datalength++;
		}

		datalength -= sharedbytes; // no need to copy the shared bytes

		// now we have decided on the block to copy, so write the
		// starting position and data length...

		tval = startloc; // initialize temp value so it can be modified

		for(c=0;c<4;c++)
		{
			onechar = tval % 0x100; // get LSByte
			fputc(onechar, outputpatch); // write it to the patch
			tval /= 0x100; // move up 1 byte
		}

		tval = datalength; // initialize temp value so it can be modified

		for(c=0;c<2;c++)
		{
			onechar = tval % 0x100; // get LSByte
			fputc(onechar, outputpatch); // write it to the patch
			tval /= 0x100; // move up one byte
		}

		// position modified rom at startloc
		fseek(modifiedrom, startloc, SEEK_SET);

		// now write the data block

		tval = datalength; // temp value etc...

		while(tval--)
		{
			onechar = fgetc(modifiedrom);
			fputc(onechar, outputpatch);
		}

		// NOW  !!!!   increment startloc appropriately (dumb dumb)
		startloc += (datalength + sharedbytes);
	}

	fclose(modifiedrom);
	fclose(outputpatch);

	printf("Success!  Patch file written to %s.\n", patchname);
	delete[] patchname; // finally deallocate the patch name space

	return(0);
}


long int filesize(FILE *stream)
{
	// finds and returns the file size of a file pointed to by stream
	long int curpos, length;

	curpos = ftell(stream);
	fseek(stream, 0L, SEEK_END);
	length = ftell(stream);
	fseek(stream, curpos, SEEK_SET);
	return length;
}

